import java.io.IOException;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.locks.*;

final class Brain
{
static int baseNg = 0, it = 0;
	static AtomicInteger nodeCnt = new AtomicInteger(0);

	static int getNodesProcessedCnt() {
		return nodeCnt.get();
	}

	static int countTypePiecesAroundXY(board b, int x, int y, cross type)
	{
		final int dimm = b.getDimensions() - 1;
		int cnt = 0;

		if (x > 0     && b.getAt(x - 1, y) == type)
			cnt++;

		if (x < dimm  && b.getAt(x + 1, y) == type)
			cnt++;

		if (y > 0     && b.getAt(x, y - 1) == type)
			cnt++;

		if (y < dimm  && b.getAt(x, y + 1) == type)
			cnt++;

		return cnt;
	}

	static List<chainpiece> getTypePiecesAroundXY(board b, int x, int y, cross type)
	{
		assert x>=0 && y>=0 && x<b.getDimensions() && y<b.getDimensions();
		List<chainpiece> types = new ArrayList<chainpiece>();

		if (x > 0                        && b.getAt(x - 1, y) == type)
			types.add(new chainpiece(x - 1, y));
		if (x < (b.getDimensions() - 1)  && b.getAt(x + 1, y) == type)
			types.add(new chainpiece(x + 1, y));
		if (y > 0                        && b.getAt(x, y - 1) == type)
			types.add(new chainpiece(x, y - 1));
		if (y < (b.getDimensions() - 1)  && b.getAt(x, y + 1) == type)
			types.add(new chainpiece(x, y + 1));

		return types;
	}

	static boolean isValidMove(board b, List<chain> chains, int x, int y, cross my_type)
	{
		assert my_type == cross.white || my_type == cross.black;

		if (b.getAt(x, y) != cross.empty)
			return false;

		// get a list of freedoms
		final int nFreedomsForThisPiece = countFreedomsAtLocation(b, x, y);

		// a piece must always have at least one freedom, else it would
		// be suicide which is not allowed
		if (nFreedomsForThisPiece > 0)
		{
//			System.out.println("" + x + "," + y + " has  > 0 freedoms");
			return true;
		}

		final player my_color = my_type == cross.white ? player.white : player.black;

		for(chain current_chain : getConnectedChains(chains, x, y, my_color))
		{
			// yes, this chain connects to this new piece
			// now verify if it still has at least 1 freedom after this connect

			// original number of freedoms for this chain
			int current_chain_n_freedoms = current_chain.getFreedoms().size();
			// one less as one freedom is used to connect to the new piece
			current_chain_n_freedoms--;
			// the new piece adds new freedoms
			current_chain_n_freedoms += nFreedomsForThisPiece;
			// if we still have at least one freedom, all is fine
// System.out.println("chain "+current_chain.getList().get(0).getX()+","+current_chain.getList().get(0).getY()+" connecting to " + x +","+y+" has " + current_chain_n_freedoms +" freedoms");
			if (current_chain_n_freedoms >= 1)
				return true;
		}

		// check if one of the 4 chains connected to x,y is opponent and if opponent and it only has one
		// freedom, then still put as it'll kill that opponent chain
		for(chain opponentChains : getConnectedChains(chains, x, y, Stop.opponentColor(my_color)))
		{
			if (opponentChains.getFreedoms().size() == 1)
			{
				// String dummy = "";
				// for(chainpiece c : opponentChains.getFreedoms())
				// 	dummy += "" + c.getX() + "," + c.getY() + " ";
				// Statics.log(" opponent, one freedom, " + opponentChains.dump());
				// Statics.log(" - freedoms " + dummy);
				return true;
			}
		}

		return false;
	}

	static List<chainpiece> getFreedomsAtLocation(board b, int x, int y)
	{
		return getTypePiecesAroundXY(b, x, y, cross.empty);
	}

	static int countFreedomsAtLocation(board b, int x, int y)
	{
		return countTypePiecesAroundXY(b, x, y, cross.empty);
	}

	static List<chain> getConnectedChains(List<chain> chains, int x, int y, player whom)
	{
		List<chain> newChains = new ArrayList<chain>();

		for(chain current_chain : chains)
		{
			if (current_chain.getColor() != whom)
				continue;

			if (current_chain.doesPieceConnectToChain(x, y))
				newChains.add(current_chain);
		}

		return newChains;
	}

	static List<chainpiece> getVictims(board b, List<chain> chains, player killerColor, chain killerChain)
	{
		List<chainpiece> victims = new ArrayList<chainpiece>();
		player victimColor = killerColor == player.white ? player.black : player.white;

		for(chain currentVictimChain : chains)
		{
			boolean connects = false;

			if (currentVictimChain.getColor() != victimColor)
				continue;

			for(chainpiece kcCurrent : killerChain.getList())
			{
				/* is this victim-chain attacked by kCurrent? */
				if (currentVictimChain.doesPieceConnectToChain(kcCurrent.getX(), kcCurrent.getY()) &&
						currentVictimChain.getFreedoms().size() == 1)
				{
					connects = true;
					break;
				}
			}

			/* then add victim-chain elements to victims-list */
			if (connects)
			{
				for(chainpiece currentVictimChainElement : currentVictimChain.getList())
					chain.addToChainpieceListKeepUnique(victims, currentVictimChainElement.getX(), currentVictimChainElement.getY());
			}
		}

		return victims;
	}

	static void setSelectedMove(SelectedMove [][] selectedMoves, int x, int y, double value, String adder)
	{
		assert adder != null;

		selectedMoves[x][y].addToValue(adder, value);
	}

	// go through opponent chains
	// divide its chain-size by its freedom size so the large the chain and the less freedoms, the sooner place a
	// stone one of the freedom-places of this chain
	static void brain_ChainSize_FreedomsSize(SelectedMove [][] selectedMoves, board b, List<chain> chains, player my_color)
	{
		MaxField mfa = new MaxField(b);
		MaxField mfb = new MaxField(b);
		MaxField mfc = new MaxField(b);

		for(chain current_chain: chains)
		{
			if (current_chain.getColor() == my_color)
				continue;

			final List<chainpiece> selected_freedoms = current_chain.getFreedoms();

			final double current_factor = (double)current_chain.size() / (double)selected_freedoms.size();
			// from this list of freedoms, pick a valid one
			for(chainpiece current_selected_freedom : selected_freedoms)
			{
				if (selectedMoves[current_selected_freedom.getX()][current_selected_freedom.getY()].isValidMove() == true)
				{
					mfa.set(current_selected_freedom.getX(), current_selected_freedom.getY(), current_factor);
					mfb.set(current_selected_freedom.getX(), current_selected_freedom.getY(), (double)current_chain.size());
					mfc.set(current_selected_freedom.getX(), current_selected_freedom.getY(), 1.0 / (double)selected_freedoms.size());
				}
			}
		}

		mfa.commit(selectedMoves, "f1a");
		mfb.commit(selectedMoves, "f1b");
		mfc.commit(selectedMoves, "f1c");
	}

	static int countByColor(board b, cross what) {
		int n = 0;

		for(int index=0; index<b.getNFields(); index++) {
			if (b.getAt(index) == what)
				n++;
		}

		return n;
	}

	// add some randomization to each field
	static void brain_Random(SelectedMove [][] selectedMoves, board b, List<chain> chains, player my_color) {
		MaxField mf = new MaxField(b);

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove())
					mf.set(x, y, Stop.genRandomDouble());
			}
		}

		mf.commit(selectedMoves, "f2");
	}

	// see if a move gives a chain of mine more freedoms
	// if a freedom-place has 1 freedom : placing a stone would reduce this chain's freedoms to zero
	//                        2 freedoms: placing a stone would keep the number of freedoms equal
	// so that's why 2 is substracted from the number of freedoms
	static void brain_Protect_my_chains(SelectedMove [][] selectedMoves, List<chain> chains, board b, player myColor) {
		MaxField mf = new MaxField(b);

		for(chain current_chain: chains) {
			if (current_chain.getColor() != myColor)
				continue;

			final List<chainpiece> currentChainNFreedoms = current_chain.getFreedoms();

			for(chainpiece currentChainFreedom : currentChainNFreedoms) {
				final int x = currentChainFreedom.getX();
				final int y = currentChainFreedom.getY();

				if (selectedMoves[x][y].isValidMove() == false)
					continue;

				int freedomLocationgetFreedomsAtLocation = getFreedomsAtLocation(b, x, y).size();

				// te first two checks are to promote creating extra eyes on a chain
				if (x > 0 && b.getAt(x - 1, y) == cross.empty &&
						x < (b.getDimensions() - 1) && b.getAt(x + 1, y) == cross.empty)
					mf.set(x, y, (double)(freedomLocationgetFreedomsAtLocation - 2) + 0.1);
				else if (y > 0 && b.getAt(x, y - 1) == cross.empty &&
						y < (b.getDimensions() - 1) && b.getAt(x, y + 1) == cross.empty)
					mf.set(x, y, (double)(freedomLocationgetFreedomsAtLocation - 2) + 0.1);
				else
					mf.set(x, y, (double)(freedomLocationgetFreedomsAtLocation - 2));
			}
		}

		mf.commit(selectedMoves, "f3");
	}

	// moves that maximum limits the number of freedoms of opponent
	static void brain_ReduceOpponentFreedoms(SelectedMove [][] selectedMoves, board b, List<chain> chains, player my_color) {
		final cross hisObj = my_color == player.white ? cross.black : cross.white;

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove() == true) {
					// count the number of pieces we are going to limit the freedomcount on
					final int curNFreedomsLimited = b.countObjectsAroundXYOfType(x, y, hisObj);

					if (curNFreedomsLimited != 0)
						setSelectedMove(selectedMoves, x, y, (double)curNFreedomsLimited, "f4");
				}
			}
		}
	}

	static void brain_FillLargestEmptySpace(SelectedMove [][] selectedMoves, board b, List<chain> chains, player my_color) {
		// ga door alle chains en bereken mijn gemiddelde afstand tot een punt en die van hem
		// dan mf (geen a of b) -hem +mijn
		MaxField mf = new MaxField(b);

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove() == false)
					continue;

				double value = 0.0;
				for(chain cur : chains) {
					double avg = 0.0;

					for(chainpiece curpiece : cur.getList()) {
						final double diffx = curpiece.getX() - x, diffy = curpiece.getY() - y;
						final double vectorLength = Math.sqrt((diffx * diffx) + (diffy * diffy));
						avg += vectorLength;
					}

					avg /= (double)cur.getList().size();

					if (cur.getColor() == my_color)
						value += avg;
					else
						value -= avg;
				}

				mf.set(x, y, value);
			}
		}

		mf.commit(selectedMoves, "f22");
	}

	// symmetry is good altough it should be position-independent
	static void brain_Symmetry(SelectedMove [][] selectedMoves, board b, List<chain> chains, player my_color) {
		cross myObj = my_color == player.white ? cross.white : cross.black;
		MaxField mf = new MaxField(b);

		for(chain current_chain : chains) {
			int move;
			int width, height;
			int minX = b.getDimensions(), minY = minX;
			int maxX = 0, maxY = 0;
			boolean left = false, right  = false;
			boolean top  = false, bottom = false;

			if (current_chain.getColor() != my_color)
				continue;

			if (current_chain.size() < 3)
				continue;

			// find bounding box for current chain
			for(chainpiece cpcc : current_chain.getList()) {
				minX = Math.min(minX, cpcc.getX());
				minY = Math.min(minY, cpcc.getY());
				maxX = Math.max(maxX, cpcc.getX());
				maxY = Math.max(maxY, cpcc.getY());
			}
			width  = maxX - minX;
			height = maxY - minY;

			move = 1;
			while (minY - (height + move) >= 0) {
				for(chainpiece cpcc : current_chain.getList()) {
					int newX = cpcc.getX();
					int newY = minY - ((cpcc.getY() - minY) + move);

					if (selectedMoves[newX][newY].isValidMove() == true)
						mf.set(newX, newY, 0.5);
				}

				move++;
			}

			move = 1;
			while (minX - (width + move) >= 0) {
				for(chainpiece cpcc : current_chain.getList()) {
					int newX = minX - ((cpcc.getX() - minX) + move);
					int newY = cpcc.getY();

					if (selectedMoves[newX][newY].isValidMove() == true)
						mf.set(newX, newY, 0.5);
				}

				move++;
			}

			move = 1;
			while (maxX + (width + move) < b.getDimensions()) {
				for(chainpiece cpcc : current_chain.getList()) {
					int newX = maxX + ((cpcc.getX() - maxX) + move);
					int newY = cpcc.getY();

					if (selectedMoves[newX][newY].isValidMove() == true)
						mf.set(newX, newY, 0.5);
				}

				move++;
			}

			move = 1;
			while (maxY + (height + move) < b.getDimensions()) {
				for(chainpiece cpcc : current_chain.getList()) {
					int newX = cpcc.getX();
					int newY = maxY + ((cpcc.getY() - maxY) + move);

					if (selectedMoves[newX][newY].isValidMove() == true)
						mf.set(newX, newY, 0.5);
				}

				move++;
			}
		}

		mf.commit(selectedMoves, "f6");
	}

	static void brain_PreventMoveRepetition(SelectedMove [][] selectedMoves, board b, List<chain> chains, List<History> history, player my_color) {
		MaxField mf = new MaxField(b);
		cross myObj = my_color == player.white ? cross.white : cross.black;

		for(History cur : history) {
			if (cur.getColor() == my_color && selectedMoves[cur.getX()][cur.getY()].isValidMove() == true) {
				mf.set(cur.getX(), cur.getY(), -2.5);
			}
		}

		mf.commit(selectedMoves, "f7");
	}

	static void brain_HuntAttackers(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor) {
		MaxField mf = new MaxField(b);
		cross myObj = myColor == player.white ? cross.white : cross.black;
		player killerColor = myColor == player.white ? player.black : player.white;

		for(chain current_chain : chains) {
			if (current_chain.getColor() != killerColor)
				continue;

			final List<chainpiece> victims = getVictims(b, chains, killerColor, current_chain);

			final List<chainpiece> hisFreedoms = current_chain.getFreedoms();

			for(chainpiece current_his_freedom : hisFreedoms) {
				if (selectedMoves[current_his_freedom.getX()][current_his_freedom.getY()].isValidMove() == true)
					mf.set(current_his_freedom.getX(), current_his_freedom.getY(), (double)victims.size() / (double)b.getDimensions());
			}
		}

		mf.commit(selectedMoves, "f9");
	}

	// FIXME: doubles can be endpoints as well?
	static List<chainpiece> findEndpoints(board b, chain chain, player myColor) {
		List<chainpiece> endPoints = new ArrayList<chainpiece>();
		cross myObj = myColor == player.white ? cross.white : cross.black;

		for(chainpiece chainPiece : chain.getList()) {
			// if connected to only one of my type, then this is an end
			if (getTypePiecesAroundXY(b, chainPiece.getX(), chainPiece.getY(), myObj).size() <= 1)
				endPoints.add(chainPiece);
		}

		if (endPoints.size() > 0)
			return endPoints;

		return null;
	}

	static void brain_extendChainsEndpoints(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor) {
		int dim = b.getDimensions();
		MaxField mf = new MaxField(b);
		cross myObj = myColor == player.white ? cross.white : cross.black;

		for(chain current_chain : chains) {
			if (current_chain.size() < 2)
				continue;

			double value = 1.0;
			player chainColor = current_chain.getColor();
			if (chainColor != myColor)
				value = 0.5;

			cross chainInChain = null;
			if (chainColor == player.white)
				chainInChain = cross.white;
			else
				chainInChain = cross.black;

			final List<chainpiece> endPoints = findEndpoints(b, current_chain, myColor);
			if (endPoints == null) // circle?
				continue;

			for(chainpiece chainPiece : endPoints) {
				List<chainpiece> startFreedoms = getFreedomsAtLocation(b, chainPiece.getX(), chainPiece.getY());
				for(chainpiece current_piece : startFreedoms) {
					int x = current_piece.getX();
					int y = current_piece.getY();

					if (selectedMoves[x][y].isValidMove() == true) {
						if (x + 2 < dim && b.getAt(x + 2, y) == chainInChain)
							value += 0.25;
						if (x - 2 >= 0 && b.getAt(x - 2, y) == chainInChain)
							value += 0.25;
						if (y + 2 < dim && b.getAt(x, y + 2) == chainInChain)
							value += 0.25;
						if (y - 2 >= 0 && b.getAt(x, y - 2) == chainInChain)
							value += 0.25;

						mf.set(x, y, value);
					}
				}
			}
		}

		mf.commit(selectedMoves, "f10");
	}

	static void brain_extendChainsAnywhere(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor) {
		MaxField mf = new MaxField(b);
		cross myObj = myColor == player.white ? cross.white : cross.black;

		for(chain current_chain : chains) {
			if (current_chain.getColor() != myColor)
				continue;

			final List<chainpiece> freedoms = current_chain.getFreedoms();

			for(chainpiece chainPiece : freedoms) {
				if (selectedMoves[chainPiece.getX()][chainPiece.getY()].isValidMove() == true) {
					final int nFreedomsAtLocation = getFreedomsAtLocation(b, chainPiece.getX(), chainPiece.getY()).size();

					mf.set(chainPiece.getX(), chainPiece.getY(), ((double)nFreedomsAtLocation / 4.0) - 0.25);
				}
			}
		}
		mf.commit(selectedMoves, "f10.2");
	}

	static void brain_preventRetakingPurge(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor, List<chainpiece> whiteCaught, List<chainpiece> blackCaught) {
		MaxField mfa = new MaxField(b);
		MaxField mfb = new MaxField(b);

		Statics.log("caught, white: " + whiteCaught.size() + ", black: " + blackCaught.size());

		// whiteCaught: objects of white which were caught
		// now check what crosses were i had stones were caught
		// then, do not (re-)use those
		for(chainpiece current : ((myColor == player.white) ? whiteCaught : blackCaught)) {
			if (selectedMoves[current.getX()][current.getY()].isValidMove() == true)
				mfa.set(current.getX(), current.getY(), -9.0);
		}

		// see what stones/crosses of the opponent i caught
		// try not to re-use those as well FIXME or should we? to really make it ours?
		for(chainpiece current : myColor == player.white ? blackCaught : whiteCaught)
		{
			if (selectedMoves[current.getX()][current.getY()].isValidMove() == true)
				mfb.set(current.getX(), current.getY(), -9.0 * 0.95);
		}

		mfa.commit(selectedMoves, "f11a");
		mfb.commit(selectedMoves, "f11b");
	}

	// return true if already in list
	static boolean addUniqueToChainPieceList(List<chainpiece> list, int x, int y) {
		for(chainpiece current : list) {
			if (current.getX() == x && current.getY() == y)
				return true;
		}

		list.add(new chainpiece(x, y));

		return false;
	}

	// return true if found in list
	static boolean removeFromChainPieceList(List<chainpiece> list, int x, int y) {
		for(int index=0; index<list.size(); index++) {
			if (list.get(index).getX() == x && list.get(index).getY() == y) {
				list.remove(index);

				return true;
			}
		}

		return false;
	}

	// FIXME also make more elaborate de-tours
	static List<chainpiece> pathBetweenPoints(board b, player myColor, int x1, int y1, int x2, int y2) {
		cross hisObj = myColor == player.white ? cross.black : cross.white;
		List<chainpiece> route = new ArrayList<chainpiece>();
		final int maxSteps = b.getDimensions() * b.getDimensions() + 1;
		int stepsTaken = 0;
		double errorX = 0.0, errorY = 0.0;

		// Statics.log("" + x1 + "," + y1 + " -> " + x2 + "," + y2);

		if (x1 == x2 && y1 == y2) {
			// Statics.log("xy1==xy2");
			return null;
		}

		// String path = "Path: ";
		do {
			int dx = x2 - x1;
			int dy = y2 - y1;
			int moveX = 0, moveY = 0;

			assert dx != 0 || dy != 0;

			if (errorX >= 1.0) {
				moveX = ((dx > 0) ? 1 : -1);
				errorX -= 1.0;
			}
			else if (errorY >= 1.0) {
				moveY = ((dy > 0) ? 1 : -1);
				errorY -= 1.0;
			}
			else {
				if (dx != 0) {
					moveX = ((dx > 0) ? 1 : -1);
					errorY += Math.abs((double)dy / (double)dx);
				}
				else {
					moveY = ((dy > 0) ? 1 : -1);
					errorX += Math.abs((double)dx / (double)dy);
				}
			}

			if (b.getAt(x1 + moveX, y1 + moveY) == hisObj) {
				// Statics.log("his obj");

				if (x1 > 0 && b.getAt(x1 - 1, y1) == cross.empty && Statics.isInThisChain(route, x1 - 1, y1) == false) {
					moveX = -1;
					moveY = 0;
				}
				else if (x1 < (b.getDimensions() - 1)&& b.getAt(x1 + 1, y1) == cross.empty && Statics.isInThisChain(route, x1 + 1, y1) == false) {
					moveX = 1;
					moveY = 0;
				}
				else if (y1 > 0 && b.getAt(x1, y1 - 1) == cross.empty && Statics.isInThisChain(route, x1, y1 - 1) == false) {
					moveX = 0;
					moveY = -1;
				}
				else if (y1 < (b.getDimensions() - 1) && b.getAt(x1, y1 + 1) == cross.empty && Statics.isInThisChain(route, x1, y1 + 1) == false) {
					moveX = 0;
					moveY = 1;
				}
				else {
					return null;
				}

				// Statics.log("Re-routed");

				errorX = errorY = 0.0;
			}

			x1 += moveX;
			y1 += moveY;
			//b.putAt(x1, y1, cross.white);
			//				Statics.logBoard(b, "R ");

			// path += "|" + x1 + "," + y1;
			if (x1 == x2 && y1 == y2) {
				// Statics.log(path);
				return route;
			}

			if (b.getAt(x1, y1) == cross.empty) {
				if (addUniqueToChainPieceList(route, x1, y1) == true) {
					// Statics.log("already in list");
					return null;
				}
			}

			stepsTaken++;
		}
		while(stepsTaken < maxSteps);

		Statics.log("too many steps");

		return null;
	}

	static void brain_CloseChain(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor) {
		MaxField mf = new MaxField(b);
		cross myObj = myColor == player.white ? cross.white : cross.black;

		for(chain current_chain : chains) {
			if (current_chain.size() < 3)
				continue;

			final List<chainpiece> endpoints = findEndpoints(b, current_chain, myColor);
			if (endpoints == null)
				continue;

			for(int index1=0; index1<(endpoints.size() - 1); index1++) {
				for(int index2=index1+1; index2<endpoints.size(); index2++) {
					final List<chainpiece> pbp = pathBetweenPoints(b, myColor, endpoints.get(index1).getX(), endpoints.get(index1).getY(), endpoints.get(index2).getX(), endpoints.get(index2).getY());

					if (pbp == null)
						continue;

					for(chainpiece point : pbp) {
						if (selectedMoves[point.getX()][point.getY()].isValidMove() == true)
							mf.set(point.getX(), point.getY(), 0.5);
					}
				}
			}
		}

		mf.commit(selectedMoves, "f12");
	}

	static boolean isValidXY(board b, int x, int y) {
		return x >= 0 && y >= 0 && x < b.getDimensions() && y < b.getDimensions();
	}

	static void cep_setMf(SelectedMove [][] selectedMoves, board b, MaxField mf, int x, int y, double value) {
		if (isValidXY(b, x, y))
			mf.set(x, y, value);
	}
			
	static void brain_ConnectEndPoints(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor) {
		MaxField mfa = new MaxField(b);
		MaxField mfb = new MaxField(b);
		cross myObj = myColor == player.white ? cross.white : cross.black;

		for(int index1=0; index1<(chains.size() - 1); index1++) {
			final List<chainpiece> endpoints1 = findEndpoints(b, chains.get(index1), myColor);

			if (endpoints1 == null)
				continue;

			for(int index2=index1+1; index2<chains.size(); index2++) {
				final List<chainpiece> endpoints2 = findEndpoints(b, chains.get(index2), myColor);

				if (endpoints2 == null)
					continue;

				for(chainpiece endpoint1 : endpoints1) {
					for(chainpiece endpoint2 : endpoints2) {
						/* path between endpoints */
						final List <chainpiece> pbp = pathBetweenPoints(b, myColor, endpoint1.getX(), endpoint1.getY(), endpoint2.getX(), endpoint2.getY());
						if (pbp == null)
							continue;

						for(chainpiece point : pbp) {
							if (selectedMoves[point.getX()][point.getY()].isValidMove() == true)
								mfa.set(point.getX(), point.getY(), 0.5);
						}

						/* prevent "knippen":
						 *   a:1wwww   e:wwwww4
						 * wwwww2            5wwww
						 *   b:3wwww   f:wwwww6
						 */
						if      (endpoint2.getX() == endpoint1.getX() + 1 && endpoint1.getY() == endpoint2.getY() + 1)	/* A */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() - 1, 0.1); /* 1 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() + 1, endpoint1.getY(), 0.1); /* 2 */
						}
						else if (endpoint2.getX() == endpoint1.getX() + 1 && endpoint1.getY() == endpoint2.getY() - 1)	/* B */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() + 1, 0.1); /* 3 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() + 1, endpoint1.getY(), 0.1); /* 2 */
						}
						else if (endpoint2.getX() == endpoint1.getX() - 1 && endpoint1.getY() == endpoint2.getY() + 1)	/* E */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() - 1, 0.1); /* 4 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() - 1, endpoint1.getY(), 0.1); /* 5 */
						}
						else if (endpoint2.getX() == endpoint1.getX() - 1 && endpoint1.getY() == endpoint2.getY() - 1)	/* F */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() + 1, 0.1); /* 6 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() - 1, endpoint1.getY(), 0.1); /* 5 */
						}
						/*  a b
						 *  w w   w 
						 *  w2w  4w6
						 *  1w3  w5w
						 *   w   w w
						 *       e f
						 */
						if      (endpoint2.getY() == endpoint1.getY() + 1 && endpoint1.getX() == endpoint2.getX() + 1)	/* F */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() + 1, endpoint1.getY(), 0.1); /* 6 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() + 1, 0.1); /* 5 */
						}
						else if (endpoint2.getY() == endpoint1.getY() + 1 && endpoint1.getY() == endpoint2.getY() - 1)	/* E */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() - 1, endpoint1.getY(), 0.1); /* 4 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() + 1, 0.1); /* 5 */
						}
						else if (endpoint2.getY() == endpoint1.getY() - 1 && endpoint1.getY() == endpoint2.getY() + 1)	/* B */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() + 1, endpoint1.getY(), 0.1); /* 3 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() - 1, 0.1); /* 2 */
						}
						else if (endpoint2.getY() == endpoint1.getY() - 1 && endpoint1.getY() == endpoint2.getY() - 1)	/* A */
						{
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX() - 1, endpoint1.getY(), 0.1); /* 1 */
							cep_setMf(selectedMoves, b, mfb, endpoint1.getX(), endpoint1.getY() - 1, 0.1); /* 2 */
						}
					}
				}
			}
		}

		mfa.commit(selectedMoves, "f13a");
		mfb.commit(selectedMoves, "f13b");
	}

	static void brain_stonesOnEmptyWithFreedoms(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor)
	{
		MaxField mf = new MaxField(b);
		cross myObj = myColor == player.white ? cross.white : cross.black;

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove() == true) {
					int freedoms = 0;

					for(chain current_chain : chains) {
						if (current_chain.doesPieceConnectToChain(x, y) == false)
							continue;

						// if connecting one is chain of my color, use its number
						// of freedoms as they're my freedoms as well
						if (current_chain.getColor() == myColor) {
							freedoms += current_chain.getFreedoms().size() - 1;				
						}
						// if the connecting object is of opponent but it has only
						// one freedom, it'll be gone anyway when this cross is
						// used so count it as 1 freedom
						else if (current_chain.getFreedoms().size() == 1) {
							freedoms++;
						}
					}

					freedoms += getFreedomsAtLocation(b, x, y).size();

					if (freedoms == 1)
						mf.set(x, y, -3.0);
					else if (freedoms == 2)
						mf.set(x, y, -2.0);
					else {
						// Statics.log("f14c: " + x + "," + y + " / " + freedoms);
						mf.set(x, y, Math.log((double)freedoms));
					}
				}
			}
		}

		mf.commit(selectedMoves, "f14");
	}

	static boolean isTypeAndValid(board b, int x, int y, cross type) {
		return isValidXY(b, x, y) && b.getAt(x, y) == type;
	}

	static boolean isEmptyAndValid(board b, int x, int y) {
		return isTypeAndValid(b, x, y, cross.empty);
	}

	static boolean isTrueIfTypeOrOutOfRange(board b, int x, int y, cross type) {
		if (isValidXY(b, x, y))
			return b.getAt(x, y) == type;

		return true;
	}

	static void brain_disallowEyes(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor)
	{
		cross myObj  = myColor == player.white ? cross.white : cross.black;
		cross hisObj = myColor == player.white ? cross.black : cross.white;

		// this method would previously only check <0,dim>
		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove() == false)
					continue;

				boolean neighboursMyColor =
					(isTrueIfTypeOrOutOfRange(b, x - 1, y, myObj) &&
					 isTrueIfTypeOrOutOfRange(b, x + 1, y, myObj) &&
					 isTrueIfTypeOrOutOfRange(b, x, y - 1, myObj) &&
					 isTrueIfTypeOrOutOfRange(b, x, y + 1, myObj));

				int nDiagonalOpponentHisColor =
					(isTypeAndValid(b, x - 1, y - 1, hisObj) ? 1 : 0) +
					(isTypeAndValid(b, x + 1, y - 1, hisObj) ? 1 : 0) +
					(isTypeAndValid(b, x - 1, y + 1, hisObj) ? 1 : 0) +
					(isTypeAndValid(b, x + 1, y + 1, hisObj) ? 1 : 0);

				if (neighboursMyColor == true && nDiagonalOpponentHisColor <= 1)
					selectedMoves[x][y].invalidate();
			}
		}
	}

	static void brain_actOnLastMove(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor, chainpiece moveThatLedToThisScene)
	{
		int x = moveThatLedToThisScene.getX();
		int y = moveThatLedToThisScene.getY();
		cross myObj  = myColor == player.white ? cross.white : cross.black;
		player hisColor = (myColor == player.white ? player.black : player.white);

		for(chain currentChain : chains) {
			if (currentChain.getColor() != hisColor)
				continue;

			if (currentChain.doesPieceConnectToChain(x, y) == false)
				continue;

			for(chainpiece currentChainPiece :currentChain.getFreedoms()) {
				if (selectedMoves[currentChainPiece.getX()][currentChainPiece.getY()].isValidMove() == true)
					setSelectedMove(selectedMoves, currentChainPiece.getX(), currentChainPiece.getY(), 0.75, "f15");
			}

			break; // should only be one
		}
	}

	static boolean tryTTRulesSub3(board b, player color, int x, int y, int dx, int dy) {
		int cx = x + dx, cy = y + dy;

		cross myCross = color == player.white ? cross.white : cross.black;

		boolean v = false;

		for(;;) {
			v = isValidXY(b, cx, cy);

			if (!v || b.getAt(cx, cy) != cross.empty)
				break;

			cx += dx;
			cy += dy;
		}

		if (v && b.getAt(cx, cy) != myCross)
			return false;

		if (!v)
			return false;

		return true;
	}

	static boolean ttRuleSub3(board b, player color, int x, int y) {
		return tryTTRulesSub3(b, color, x, y, -1, 0) &&
			tryTTRulesSub3(b, color, x, y, 1, 0) &&
			tryTTRulesSub3(b, color, x, y, 0, -1) &&
			tryTTRulesSub3(b, color, x, y, 0, 1);
	}

	// http://senseis.xmp.net/?LogicalRules
	static int countTrompTaylorPoins(board b, player color) {
		int cnt = 0;

		cross myCross = color == player.white ? cross.white : cross.black;

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				cross c = b.getAt(x, y);

				if (c == myCross)
					cnt++;
				else if (c == cross.empty && ttRuleSub3(b, color, x, y))
					cnt++;
			}
		}

		return cnt;
	}

	static void brain_TrompTaylerPointsIncrease(SelectedMove [][] selectedMoves, board b, player myColor, scene s) throws InterruptedException {
		MaxField mf = new MaxField(b);

		int cur = countTrompTaylorPoins(b, myColor);

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove() == true) {
					scene sceneAfterMove = selectedMoves[x][y].getSceneAfterMove(s, x, y, myColor);

					int then = countTrompTaylorPoins(sceneAfterMove.getBoard(), myColor);

					mf.set(x, y, then - cur);
				}
			}
		}

		mf.commit(selectedMoves, "f21");
	}

	// kijken naar de selectedMove voor zetten na deze, daar dan de sterkste voorkomen <- eventueel at random wat zetten checken (montecarlo!)
	static void brain_compareWithAfterMove_ValueOfBestMove(SelectedMove [][] selectedMoves, board b, List<chain> chains, player myColor, scene s, boolean random_allowed, double thinkTime) throws InterruptedException {
		MaxField mfa = new MaxField(b);
		MaxField mfb = new MaxField(b);
		player hisColor = (myColor == player.white) ? player.black : player.white;

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				if (selectedMoves[x][y].isValidMove() == true) {
					scene sceneAfterMove = selectedMoves[x][y].getSceneAfterMove(s, x, y, myColor);
					SelectedMove selectedMoveAfterMove = sceneAfterMove.decideMove(hisColor, random_allowed, false, thinkTime, new ArrayList<Output>(), 0);
					if (selectedMoveAfterMove != null) {
						mfa.set(x, y, -selectedMoveAfterMove.getValue());

						mfb.set(selectedMoveAfterMove.getX(), selectedMoveAfterMove.getY(), 0.1);
					}
					else {
						mfa.set(x, y, 1.0);
					}
				}
			}
		}

		mfa.commit(selectedMoves, "f20a");
		mfb.commit(selectedMoves, "f20b");
	}

	// this actually checks super KO
	// 'regular' ko: previous board position should not be repeated
	// super ko: do not allow any board position to be repeated
	static boolean isKO(List<History> history, board newBoard) {
		for(History h : history) {
			if (h.getBoard().equals(newBoard) == true)
				return true;
		}

		return false;
	}

	static class MonteCarloThread implements Runnable {
		private SelectedMove [][] selectedMoves = null;
		private Semaphore smLock = null;
		private scene s = null;
		private player startColor = null;
		private long endTs = 0;
		private MersenneTwisterFast r = new MersenneTwisterFast();

		MonteCarloThread(SelectedMove [][] selectedMoves, Semaphore smLock, scene s, player color, long endTs) {
			this.selectedMoves = selectedMoves;
			this.smLock = smLock;
			this.s = s;
			this.startColor = color;
			this.endTs = endTs;
		}

		int [] selectRandomField(SelectedMove [][] selectedMoves, scene s) {
			final board b = s.getBoard();
			final int dim = b.getDimensions();

			for(int loop=0; loop<dim; loop++) {
				int x = r.nextInt(dim);
				int y = r.nextInt(dim);

				if (selectedMoves[x][y].isValidMove() == true)
					return new int [] { x, y };
			}

			int baseX = r.nextInt(dim);
			int baseY = r.nextInt(dim);

			for(int ly=0; ly<dim; ly++) {
				for(int lx=0; lx<dim; lx++) {
					int x = (lx + baseX) % dim;
					int y = (ly + baseY) % dim;

					if (selectedMoves[x][y].isValidMove() == true)
						return new int [] { x, y };
				}
			}

			return null;
		}

		public void run() {
			final board b = s.getBoard();
			final int dim = b.getDimensions();
			MaxField mfa = new MaxField(b);

			int mcDone = 0, md = 0;
			long tDepth = 0;

			while(System.currentTimeMillis() < endTs) {
				int [] newMove = selectRandomField(selectedMoves, s);
				int [] startMove = newMove;

				if (newMove == null)
					break;

				player winner = null;
				player color = startColor;

				scene workS = null;
				int depth = 0;
				for(int loop=0; loop<dim * dim; loop++) {
					depth++;

					if (workS == null)
						workS = s.doMove(newMove[0], newMove[1], color == player.white ? cross.white : cross.black, color);
					else
						workS = workS.doMove(newMove[0], newMove[1], color == player.white ? cross.white : cross.black, color);

					player newCurrentPlayer = Stop.opponentColor(color);
					workS.initSelectedMoves(newCurrentPlayer);

					SelectedMove [][] sm = workS.getSelectedMoves(newCurrentPlayer);
					newMove = selectRandomField(sm, workS);

					if (newMove == null) {
						if (countTrompTaylorPoins(workS.getBoard(), color) > countTrompTaylorPoins(workS.getBoard(), newCurrentPlayer))
							winner = color;
						else
							winner = newCurrentPlayer;
						break;
					}

					color = newCurrentPlayer;
				}

				if (winner == color)
					mfa.add(startMove[0], startMove[1], 1);
				else if (winner != null)
					mfa.add(startMove[0], startMove[1], -1);

				if (depth > md)
					md = depth;

				tDepth += depth;

				mcDone++;
			}

			Statics.log("mcDone: " + mcDone + ", mdepth: " + md + ", adepth: " + (double)tDepth / mcDone);

			smLock.acquireUninterruptibly();
			mfa.commit(selectedMoves, "f50");
			smLock.release();
		}
	}

	static void brain_MonteCarlo(SelectedMove [][] selectedMoves, scene s, player color, long endTs) throws InterruptedException {
		//int nCpus = Runtime.getRuntime().availableProcessors();
		int nCpus = 1;

		Statics.log("Starting " + nCpus + " threads");

		Thread [] handles = new Thread[nCpus];

		Semaphore lock = new Semaphore(1, false);

		long endTsThreads = endTs - 100;

		for(int loop=0; loop<nCpus; loop++) {
			MonteCarloThread instance = new MonteCarloThread(selectedMoves, lock, s, color, endTsThreads);

			handles[loop] = new Thread(instance);
			handles[loop].start();
		}

		Statics.log("Threads started");

		for(int loop=0; loop<nCpus; loop++)
			handles[loop].join();

		Statics.log("Threads terminated");
	}

	static SelectedMove decideMove(final SelectedMove [][] selectedMoves, final scene curScene, final player myColor, boolean random_allowed, boolean deep, double thinkTime, List<Output> outputs, int moveNr) throws InterruptedException {
		long endTs = System.currentTimeMillis() + (long)(thinkTime * 1000.0);

		chainpiece moveThatLedToThisScene = curScene.getMoveThatLedToThisScene();
		board b = curScene.getBoard();
		List<chain> chains = curScene.getChains();
		List<History> history = curScene.getHistory();
		List<chainpiece> whiteCaught = curScene.getWhiteCaught();
		List<chainpiece> blackCaught = curScene.getBlackCaught();
		int stonesCapturedInMTLTTS = curScene.getStonesCapturedInMTLTTS();
		cross myObj = myColor == player.white ? cross.white : cross.black;
		int selectedMoveInc = -1, selectedMoveDec = 10000000;
		double selectedMoveValue = -(1 << 31);
		SelectedMove selectedMove = null;
		int whiteCount = b.countObjects(cross.white);
		int blackCount = b.countObjects(cross.black);
		double totalValue = 0.0, totalStdDev = 0.0;
		int nValue = 0;
		long startTs = System.currentTimeMillis();

		// System.out.println("decidemove ");
		/* prevent board being filled up with only one color, then being emptied when there's
		 * only one cross empty, etc.
		 */
		if ((myColor == player.white && whiteCount > b.getDimensions() && blackCount == 0) ||
				(myColor == player.black && blackCount > b.getDimensions() && whiteCount == 0))
		{
			Statics.log("PASS because only one color on board");
			return null;
		}

		brain_disallowEyes(selectedMoves, b, chains, myColor);
		brain_ChainSize_FreedomsSize(selectedMoves, b, chains, myColor);
		if (random_allowed)
			brain_Random(selectedMoves, b, chains, myColor);
// ATARI BEVEILIGING D'R IN FIXME
		brain_Protect_my_chains(selectedMoves, chains, b, myColor);
		brain_ReduceOpponentFreedoms(selectedMoves, b, chains, myColor);
		brain_FillLargestEmptySpace(selectedMoves, b, chains, myColor);
		brain_Symmetry(selectedMoves, b, chains, myColor);
		brain_PreventMoveRepetition(selectedMoves, b, chains, history, myColor);
		brain_HuntAttackers(selectedMoves, b, chains, myColor);
		brain_extendChainsEndpoints(selectedMoves, b, chains, myColor);
		brain_extendChainsAnywhere(selectedMoves, b, chains, myColor);
		brain_preventRetakingPurge(selectedMoves, b, chains, myColor, whiteCaught, blackCaught);
		brain_CloseChain(selectedMoves, b, chains, myColor);
		brain_ConnectEndPoints(selectedMoves, b, chains, myColor);
		brain_stonesOnEmptyWithFreedoms(selectedMoves, b, chains, myColor);
		brain_TrompTaylerPointsIncrease(selectedMoves, b, myColor, curScene);

		if (moveThatLedToThisScene != null)
			brain_actOnLastMove(selectedMoves, b, chains, myColor, moveThatLedToThisScene);

		if (deep) {
			// brain_compareWithAfterMove_CaptureCount(selectedMoves, b, chains, myColor, curScene); FIXME: maar dan alleen zinnig bij multiple zetten diep
// niet nodig voor markov2			brain_compareWithAfterMove_ValueOfBestMove(selectedMoves, b, chains, myColor, curScene, random_allowed, thinkTime); // combineer met brain_compareWithAfterMove_CaptureCount qua diep zoeken

			brain_MonteCarlo(selectedMoves, curScene, myColor, endTs);
		}

		for(int y=0; y<b.getDimensions(); y++) {
			for(int x=0; x<b.getDimensions(); x++) {
				boolean pick = false;

				if (selectedMoves[x][y].isValidMove() == false)
					continue;

				if (selectedMoves[x][y].getCount() > 0) {
					totalValue += selectedMoves[x][y].getValue();
					totalStdDev += selectedMoves[x][y].getValue() * selectedMoves[x][y].getValue();
					nValue++;
				}

				// Statics.log("sm " + x + "," + y + ": " + selectedMoves[x][y].getValue() + selectedMoves[x][y].dump());

				if (selectedMove == null)
					pick = true;
				else if (selectedMoves[x][y].getValue() > selectedMoveValue)
					pick = true;
				else if (selectedMoves[x][y].getValue() == selectedMoveValue && selectedMoves[x][y].getInc() > selectedMoveInc)
				{
					Statics.log("pick movecount");
					pick = true;
				}
				else if (selectedMoves[x][y].getValue() == selectedMoveValue && selectedMoves[x][y].getInc() == selectedMoveInc && selectedMoves[x][y].getDec() < selectedMoveDec)
				{
					Statics.log("pick moveinc");
					pick = true;
				}

				if (pick && selectedMoves[x][y].getCount() == 0)
				{
					// Statics.log("Move chosen but not chosen...");
					pick = false;
				}

				if (pick)
				{
					board boardAfterMove = selectedMoves[x][y].getSceneAfterMove(curScene, x, y, myColor).getBoard();

					if (isKO(history, boardAfterMove) == true)
					{
						Statics.log("Move not used to prevent repetition: " + x + "," + y);
						pick = false;
					}

					// if the piece placed at 'moveThatLedToThisScene' is gone after this move, it
					// seems this move captures it. Now if moveThatLedToThisScene captured only one
					// piece, this move would've been KO according to http://en.wikipedia.org/wiki/Ko_rule#Ko
					if (moveThatLedToThisScene != null &&
							boardAfterMove.getAt(moveThatLedToThisScene.getX(), moveThatLedToThisScene.getY()) == cross.empty &&
							stonesCapturedInMTLTTS == 1)
					{
						Statics.log("Other form of prevent repetition: " + x + "," + y);
						pick = false;
					}

				}

				if (pick)
				{
					selectedMove = selectedMoves[x][y];
					selectedMoveValue = selectedMoves[x][y].getValue();
					selectedMoveInc   = selectedMoves[x][y].getInc();
					selectedMoveDec   = selectedMoves[x][y].getDec();
				}
			}
		}

		if (nValue > 0)
		{
			double avg = totalValue / (double)nValue;
			double stddev = Math.sqrt((totalStdDev / (double)nValue) - Math.pow(avg, 2.0));

			Statics.log("" + myColor + " Average value: " + avg + ", std.dev.: " + stddev + ", # values: " + nValue);
		}

		if (selectedMove == null)
			Statics.log("" + myColor + " No moves selected");
		else
			Statics.log("" + myColor + " Selected: " + selectedMove.getX() + "," + selectedMove.getY() + " " + selectedMoves[selectedMove.getX()][selectedMove.getY()].dump());

		for(Output o : outputs) {
			try {
				chainpiece cp = null;
				if (selectedMove != null)
					cp = new chainpiece(selectedMove.getX(), selectedMove.getY());
				o.emitBoard(moveNr, b, selectedMoves, cp, myColor);
			}
			catch(IOException ie) {
				Statics.log("Failed to emit board: " + ie);
			}
		}

		return selectedMove;
	}
}
